#pragma once

#include <openssl/x509v3.h>
#include <openssl/ssl.h>

inline X509_OBJECT *
X509_OBJECT_new()
{
  return (X509_OBJECT *) calloc(1, sizeof(X509_OBJECT));
}

inline void
X509_OBJECT_free(X509_OBJECT *a)
{
  X509_OBJECT_free_contents(a);
  free(a);
}

#undef X509_LU_X509
#undef X509_LU_CRL
enum X509_LOOKUP_TYPE
  {
    X509_LU_NONE = 0,
    X509_LU_X509,
    X509_LU_CRL
  };

inline X509_OBJECT *
X509_STORE_CTX_get_obj_by_subject(X509_STORE_CTX *vs,
                                  X509_LOOKUP_TYPE type,
                                  X509_NAME *name)
{
  X509_OBJECT *ret = X509_OBJECT_new();
  if (!X509_STORE_get_by_subject(vs, type, name, ret))
    {
      X509_OBJECT_free(ret);
      return nullptr;
    }

  return ret;
}

inline X509_CRL *
X509_OBJECT_get0_X509_CRL(X509_OBJECT *a)
{
  if (a == nullptr || a->type != X509_LU_CRL)
    return nullptr;
  return a->data.crl;
}

inline const ASN1_TIME *
X509_CRL_get0_lastUpdate(const X509_CRL *crl)
{
  return crl->crl->lastUpdate;
}

inline const ASN1_TIME *
X509_CRL_get0_nextUpdate(const X509_CRL *crl)
{
  return crl->crl->nextUpdate;
}

inline const ASN1_INTEGER *
X509_REVOKED_get0_serialNumber(const X509_REVOKED *x)
{
  return x->serialNumber;
}

inline ASN1_INTEGER *
X509_REVOKED_get0_serialNumber(X509_REVOKED *x)
{
  return x->serialNumber;
}

inline const ASN1_INTEGER *
X509_get0_serialNumber(const X509 *a)
{
  return a->cert_info->serialNumber;
}

inline int
SSL_up_ref(SSL *s)
{
  CRYPTO_add(&s->references, 1, CRYPTO_LOCK_SSL);
  return 1;
}

inline int
BIO_get_shutdown(BIO *a)
{
  return a->shutdown;
}

inline void
BIO_set_shutdown(BIO *a, int shut)
{
  a->shutdown = shut;
}

inline void
BIO_set_data(BIO *a, void *ptr)
{
  a->ptr = ptr;
}

inline void
BIO_set_init(BIO *a, int init)
{
  a->init = init;
}

inline BIO_METHOD *
BIO_meth_new(int type, const char *name)
{
  BIO_METHOD *biom = (BIO_METHOD *) calloc(1, sizeof(BIO_METHOD));
  biom->type = type;
  biom->name = name;
  return biom;
}

inline int
BIO_meth_set_write(BIO_METHOD *biom,
                   int (*bwrite) (BIO *, const char *, int))
{
  biom->bwrite = bwrite;
  return 1;
}

inline int
BIO_meth_set_read(BIO_METHOD *biom,
                  int (*bread) (BIO *, char *, int))
{
  biom->bread = bread;
  return 1;
}

inline int
BIO_meth_set_puts(BIO_METHOD *biom,
                  int (*bputs) (BIO *, const char *))
{
  biom->bputs = bputs;
  return 1;
}


inline int
BIO_meth_set_ctrl(BIO_METHOD *biom,
                  long (*ctrl) (BIO *, int, long, void *))
{
  biom->ctrl = ctrl;
  return 1;
}

inline int
BIO_meth_set_create(BIO_METHOD *biom, int (*create) (BIO *))
{
  biom->create = create;
  return 1;
}

inline int
BIO_meth_set_destroy(BIO_METHOD *biom, int (*destroy) (BIO *))
{
  biom->destroy = destroy;
  return 1;
}

inline void *
BIO_get_data(BIO *a)
{
  return a->ptr;
}

inline int
X509_REQ_get_signature_nid(const X509_REQ *req)
{
  return OBJ_obj2nid(req->sig_alg->algorithm);
}

inline X509_PUBKEY *
X509_REQ_get_X509_PUBKEY(X509_REQ *req)
{
  return req->req_info->pubkey;
}

inline const ASN1_TIME *
X509_REVOKED_get0_revocationDate(const X509_REVOKED *x)
{
  return x->revocationDate;
}

inline int
X509_CRL_set1_lastUpdate(X509_CRL *x, ASN1_TIME *tm)
{
  x->crl->lastUpdate = tm;
  return 1;
}

inline int
X509_CRL_set1_nextUpdate(X509_CRL *x, ASN1_TIME *tm)
{
  x->crl->nextUpdate = tm;
  return 1;
}

inline int
X509_CRL_get_signature_nid(const X509_CRL *crl)
{
    return OBJ_obj2nid(crl->sig_alg->algorithm);
}

inline const STACK_OF(X509_EXTENSION) *
X509_get0_extensions(const X509 *x)
{
  return x->cert_info->extensions;
}

inline const STACK_OF(X509_EXTENSION) *
X509_CRL_get0_extensions(const X509_CRL *crl)
{
  return crl->crl->extensions;
}

inline int
X509_up_ref(X509 *x)
{
  CRYPTO_add(&x->references, 1, CRYPTO_LOCK_X509);
  return 1;
}

inline int
X509_CRL_up_ref(X509_CRL *crl)
{
  CRYPTO_add(&crl->references,1,CRYPTO_LOCK_X509_CRL);
  return 1;
}

inline unsigned char *
ASN1_STRING_get0_data(ASN1_STRING *x)
{
  return ASN1_STRING_data(x);
}

inline X509 *
X509_STORE_CTX_get0_cert(X509_STORE_CTX *ctx)
{
  return ctx->cert;
}

inline int
X509_OBJECT_get_type(const X509_OBJECT *a)
{
  return a->type;
}

inline int
EVP_PKEY_up_ref(EVP_PKEY *pkey)
{
  CRYPTO_add(&pkey->references, 1, CRYPTO_LOCK_EVP_PKEY);
  return 1;
}

inline const SSL_METHOD *
TLS_server_method()
{
  return SSLv23_server_method();
}

inline const SSL_METHOD *
TLS_client_method()
{
  return SSLv23_client_method();
}

inline HMAC_CTX *
HMAC_CTX_new(void)
{
  HMAC_CTX *ctx = (HMAC_CTX *) calloc(1, sizeof(HMAC_CTX));
  HMAC_CTX_init(ctx);
  return ctx;
}

inline void
HMAC_CTX_free(HMAC_CTX *ctx)
{
  HMAC_CTX_cleanup(ctx);
  free(ctx);
}

inline size_t
OBJ_length(const ASN1_OBJECT *obj)
{
  return obj->length;
}

inline int
RSA_bits(const RSA *r)
{
  return (BN_num_bits(r->n));
}

inline int
RSA_set0_key(RSA *r, BIGNUM *n, BIGNUM *e, BIGNUM *d)
{
  if (n)
    {
      BN_free(r->n);
      r->n = n;
    }
  if (e)
    {
      BN_free(r->e);
      r->e = e;
    }
  if (d)
    {
      BN_free(r->d);
      r->d = d;
    }

  return 1;
}

inline void
RSA_get0_key(const RSA *r,
             const BIGNUM **n, const BIGNUM **e, const BIGNUM **d)
{
  if (n)
    *n = r->n;
  if (e)
    *e = r->e;
  if (d)
    *d = r->d;
}

inline int
RSA_set0_factors(RSA *r, BIGNUM *p, BIGNUM *q)
{
  if (p)
    {
      BN_free(r->p);
      r->p = p;
    }
  if (q)
    {
      BN_free(r->q);
      r->q = q;
    }

  return 1;
}

inline int
RSA_set0_crt_params(RSA *r, BIGNUM *dmp1, BIGNUM *dmq1, BIGNUM *iqmp)
{
  if (dmp1)
    {
      BN_free(r->dmp1);
      r->dmp1 = dmp1;
    }
  if (dmq1)
    {
      BN_free(r->dmq1);
      r->dmq1 = dmq1;
    }
  if (iqmp)
    {
      BN_free(r->iqmp);
      r->iqmp = iqmp;
    }

  return 1;
}

inline int
DSA_set0_pqg(DSA *d, BIGNUM *p, BIGNUM *q, BIGNUM *g)
{
  if (p)
    {
      BN_free(d->p);
      d->p = p;
    }
  if (q)
    {
      BN_free(d->q);
      d->q = q;
    }
  if (g)
    {
      BN_free(d->g);
      d->g = g;
    }

  return 1;
}

inline int
DSA_set0_key(DSA *d, BIGNUM *pub_key, BIGNUM *priv_key)
{
  if (pub_key)
    {
      BN_free(d->pub_key);
      d->pub_key = pub_key;
    }
  if (priv_key)
    {
      BN_free(d->priv_key);
      d->priv_key = priv_key;
    }

  return 1;
}

inline void
DSA_SIG_get0(const DSA_SIG *sig, const BIGNUM **pr, const BIGNUM **ps)
{
  if (pr)
    *pr = sig->r;
  if (ps)
    *ps = sig->s;
}

inline int
DSA_SIG_set0(DSA_SIG *sig, BIGNUM *r, BIGNUM *s)
{
  BN_clear_free(sig->r);
  BN_clear_free(sig->s);
  sig->r = r;
  sig->s = s;
  return 1;
}

inline void
DSA_get0_pqg(const DSA *d,
             const BIGNUM **p, const BIGNUM **q, const BIGNUM **g)
{
  if (p)
    *p = d->p;
  if (q)
    *q = d->q;
  if (g)
    *g = d->g;
}

inline void
DSA_get0_key(const DSA *d,
             const BIGNUM **pub_key, const BIGNUM **priv_key)
{
  if (pub_key)
    *pub_key = d->pub_key;
  if (priv_key)
    *priv_key = d->priv_key;
}

inline RSA *
EVP_PKEY_get0_RSA(EVP_PKEY *pkey)
{
  return pkey->pkey.rsa;
}

inline DSA *
EVP_PKEY_get0_DSA(EVP_PKEY *pkey)
{
  return pkey->pkey.dsa;
}

inline int
DH_set0_pqg(DH *dh, BIGNUM *p, BIGNUM *q, BIGNUM *g)
{
  if (p)
    {
      BN_free(dh->p);
      dh->p = p;
    }
  if (q)
    {
      BN_free(dh->q);
      dh->q = q;
    }
  if (g)
    {
      BN_free(dh->g);
      dh->g = g;
    }

  if (q) {
    dh->length = BN_num_bits(q);
  }

  return 1;
}

inline void
DH_get0_pqg(const DH *dh,
            const BIGNUM **p, const BIGNUM **q, const BIGNUM **g)
{
  if (p)
    *p = dh->p;
  if (q)
    *q = dh->q;
  if (g)
    *g = dh->g;
}

inline void
DH_get0_key(const DH *dh, const BIGNUM **pub_key, const BIGNUM **priv_key)
{
  if (pub_key)
    *pub_key = dh->pub_key;
  if (priv_key)
    *priv_key = dh->priv_key;
}

inline EVP_CIPHER *
EVP_CIPHER_meth_new(int cipher_type, int block_size, int key_len)
{
  EVP_CIPHER *cipher = (EVP_CIPHER *) calloc(1, sizeof(EVP_CIPHER));
  if (cipher)
    {
      cipher->nid = cipher_type;
      cipher->block_size = block_size;
      cipher->key_len = key_len;
    }

  return cipher;
}

inline void
EVP_CIPHER_meth_free(EVP_CIPHER *cipher)
{
  free(cipher);
}

inline int
EVP_CIPHER_meth_set_iv_length(EVP_CIPHER *cipher, int iv_len)
{
  cipher->iv_len = iv_len;
  return 1;
}

inline int
EVP_CIPHER_meth_set_flags(EVP_CIPHER *cipher, unsigned long flags)
{
  cipher->flags = flags;
  return 1;
}

inline int
EVP_CIPHER_meth_set_init(EVP_CIPHER *cipher,
                         int (*init) (EVP_CIPHER_CTX *ctx,
                                      const unsigned char *key,
                                      const unsigned char *iv,
                                      int enc))
{
  cipher->init = init;
  return 1;
}

inline int
EVP_CIPHER_meth_set_do_cipher(EVP_CIPHER *cipher,
                              int (*do_cipher) (EVP_CIPHER_CTX *ctx,
                                                unsigned char *out,
                                                const unsigned char *in,
                                                size_t inl))
{
  cipher->do_cipher = do_cipher;
  return 1;
}

inline int
EVP_CIPHER_meth_set_cleanup(EVP_CIPHER *cipher,
                            int (*cleanup) (EVP_CIPHER_CTX *))
{
  cipher->cleanup = cleanup;
  return 1;
}

inline int
EVP_CIPHER_meth_set_impl_ctx_size(EVP_CIPHER *cipher, int ctx_size)
{
  cipher->ctx_size = ctx_size;
  return 1;
}

inline int
EVP_CIPHER_meth_set_set_asn1_params(EVP_CIPHER *cipher,
                                    int (*set_asn1_parameters) (EVP_CIPHER_CTX *,
                                                                ASN1_TYPE *))
{
  cipher->set_asn1_parameters = set_asn1_parameters;
  return 1;
}

inline int
EVP_CIPHER_meth_set_get_asn1_params(EVP_CIPHER *cipher,
                                    int (*get_asn1_parameters) (EVP_CIPHER_CTX *,
                                                                ASN1_TYPE *))
{
  cipher->get_asn1_parameters = get_asn1_parameters;
  return 1;
}

inline int
EVP_CIPHER_meth_set_ctrl(EVP_CIPHER *cipher,
                         int (*ctrl) (EVP_CIPHER_CTX *, int type,
                                      int arg, void *ptr))
{
  cipher->ctrl = ctrl;
  return 1;
}
